#include <msp/core/environ.h>
#include <msp/fs/dir.h>
#include <msp/fs/stat.h>
#include <msp/strings/format.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include "androidarchiver.h"
#include "androidassetpackagingtool.h"
#include "androidcompiler.h"
#include "androidlinker.h"
#include "androidmanifestgenerator.h"
#include "androidtools.h"
#include "apkbuilder.h"
#include "architecture.h"
#include "builder.h"
#include "jarsigner.h"

using namespace std;
using namespace Msp;

// TODO Mark problems somewhere instead of throwing exceptions

unsigned parse_version(const std::string &version_str)
{
	vector<string> version_parts = split(version_str, '.');
	unsigned version = lexical_cast<unsigned>(version_parts[0])<<16;
	if(version_parts.size()>1)
		version += lexical_cast<unsigned>(version_parts[1])<<8;
	if(version_parts.size()>2)
		version += lexical_cast<unsigned>(version_parts[2]);
	return version;
}


AndroidDevKit::AndroidDevKit(Builder &b, const string &type, const FS::Path &default_path):
	builder(b)
{
	string var = format("ANDROID_%s_ROOT", type);
	root = getenv(var);
	if(root.empty())
	{
		if(!default_path.empty() && FS::exists(default_path))
			root = default_path;
		else
		{
			builder.get_logger().log("problems", format("Android %s not found", type));
			return;
		}
	}

	FS::Path platforms_dir = root/"platforms";
	if(!FS::exists(platforms_dir))
		return;

	builder.get_logger().log("files", format("Traversing %s", platforms_dir.str()));
	vector<string> platforms = list_filtered(platforms_dir, "^android-[1-9][0-9]*$");

	for(vector<string>::const_iterator i=platforms.begin(); i!=platforms.end(); ++i)
		supported_api_levels.insert(lexical_cast<unsigned>(i->substr(8)));
}

void AndroidDevKit::select_api_level(unsigned api)
{
	if(!supported_api_levels.count(api))
		throw invalid_argument("AndroidDevKit::select_api_level");

	init_api_level(api);
}


AndroidSdk::AndroidSdk(Builder &b):
	AndroidDevKit(b, "SDK")
{
	find_build_tools_dir();
}

void AndroidSdk::find_build_tools_dir()
{
	FS::Path bt_dir = root/"build-tools";
	if(!FS::exists(bt_dir))
	{
		builder.get_logger().log("problems", "Android build-tools not found");
		return;
	}

	builder.get_logger().log("files", format("Traversing %s", bt_dir.str()));
	vector<string> tool_versions = list_files(bt_dir);

	string use_tools;
	unsigned latest_version = 0;
	for(vector<string>::const_iterator i=tool_versions.begin(); i!=tool_versions.end(); ++i)
	{
		unsigned version = parse_version(*i);
		if(version>latest_version)
		{
			use_tools = *i;
			latest_version = version;
		}
	}

	if(use_tools.empty())
	{
		builder.get_logger().log("problems", "Android build-tools not found");
		return;
	}

	build_tools_dir = bt_dir/use_tools;
	builder.get_logger().log("tools", format("Android build-tools found in %s", build_tools_dir.str()));
}

void AndroidSdk::init_api_level(unsigned api)
{
	platform_jar = root/"platforms"/format("android-%d", api)/"android.jar";
}


AndroidNdk::AndroidNdk(Builder &b, const Architecture &a, const AndroidSdk &sdk):
	AndroidDevKit(b, "NDK", create_default_path(sdk)),
	architecture(a)
{
	if(!root.empty())
	{
		FS::Path csr = root/"sysroot";
		if(FS::exists(csr))
		{
			common_sysroot = csr;
			builder.get_logger().log("tools", format("Android NDK common sysroot is %s", common_sysroot));
		}
	}

	find_toolchain_dir();
}

FS::Path AndroidNdk::create_default_path(const AndroidSdk &sdk)
{
	if(sdk.get_root_dir().empty())
		return FS::Path();
	return sdk.get_root_dir()/"ndk-bundle";
}

void AndroidNdk::find_toolchain_dir()
{
	if(root.empty())
		return;

	FS::Path toolchains_dir = root/"toolchains";
	if(!FS::exists(toolchains_dir))
	{
		builder.get_logger().log("problems", "Android NDK toolchains not found");
		return;
	}

	builder.get_logger().log("files", format("Traversing %s", toolchains_dir.str()));
	string prefix = architecture.get_cross_prefix()+"-";
	vector<string> toolchains = list_filtered(toolchains_dir, "^"+prefix);

	string use_toolchain;
	unsigned latest_version = 0;
	for(vector<string>::const_iterator i=toolchains.begin(); i!=toolchains.end(); ++i)
	{
		string version_str = i->substr(prefix.size());
		string compiler = "gcc";
		if(!isdigit(version_str[0]))
		{
			unsigned j;
			for(j=1; (j<version_str.size() && !isdigit(version_str[j])); ++j) ;
			compiler = version_str.substr(0, j);
			version_str = version_str.substr(j);
		}

		if(compiler!="gcc")
			continue;

		unsigned version = parse_version(version_str);
		if(version>latest_version)
		{
			use_toolchain = *i;
			latest_version = version;
		}
	}

	if(use_toolchain.empty())
	{
		builder.get_logger().log("problems", "No applicable Android NDK toolchains found");
		return;
	}

	const Architecture &native_arch = builder.get_native_arch();

	FS::Path tc_archs_dir = toolchains_dir/use_toolchain/"prebuilt";
	builder.get_logger().log("files", format("Traversing %s", tc_archs_dir.str()));
	string use_arch = native_arch.best_match(list_files(tc_archs_dir));

	if(use_arch.empty())
	{
		builder.get_logger().log("problems", "No applicable Android NDK toolchains found");
		return;
	}

	bin_dir = toolchains_dir/use_toolchain/"prebuilt"/use_arch/"bin";
	builder.get_logger().log("tools", format("Android NDK toolchain binaries found in %s", bin_dir.str()));
}

void AndroidNdk::init_api_level(unsigned api)
{
	FS::Path platform_archs_dir = root/"platforms"/format("android-%d", api);
	builder.get_logger().log("files", format("Traversing %s", platform_archs_dir.str()));
	vector<string> platform_archs = list_filtered(platform_archs_dir, "^arch-");
	for(vector<string>::iterator i=platform_archs.begin(); i!=platform_archs.end(); ++i)
		i->erase(0, 5);
	string use_arch = architecture.best_match(platform_archs);

	if(use_arch.empty())
	{
		builder.get_logger().log("problems", "No matching Android NDK platform found");
		return;
	}

	platform_sysroot = platform_archs_dir/("arch-"+use_arch);
	builder.get_logger().log("tools", format("Android NDK platform sysroot is %s", platform_sysroot));
}


AndroidTools::AndroidTools(Builder &builder, const Architecture &arch):
	sdk(builder),
	ndk(builder, arch, sdk)
{
	const set<unsigned> &sdk_api_levels = sdk.get_supported_api_levels();
	const set<unsigned> &ndk_api_levels = ndk.get_supported_api_levels();
	unsigned highest_common = 0;
	for(set<unsigned>::const_reverse_iterator i=sdk_api_levels.rbegin(); (!highest_common && i!=sdk_api_levels.rend()); ++i)
		if(ndk_api_levels.count(*i))
			highest_common = *i;

	if(highest_common)
	{
		sdk.select_api_level(highest_common);
		ndk.select_api_level(highest_common);
	}
	else
		builder.get_logger().log("problems", "No usable Android platforms found");

	add_tool(new AndroidCompiler(builder, arch, "CC", ndk));
	add_tool(new AndroidCompiler(builder, arch, "CXX", ndk));
	add_tool(new AndroidLinker(builder, arch, ndk));
	add_tool(new AndroidArchiver(builder, arch, ndk));
	add_tool(new AndroidManifestGenerator(builder));
	add_tool(new AndroidAssetPackagingTool(builder, sdk));
	add_tool(new ApkBuilder(builder));
	add_tool(new JarSigner(builder));
}
